#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <stdarg.h>
#include <unistd.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>


#include "cuda_render.h"
void error_exit(const char* format, ... )
{
  va_list args;
  va_start( args, format );
  vfprintf( stderr, format, args );
  va_end( args );
  exit(1); 
}

#include "graphics_interface.c"

#define checkCUDAError() \
{ \
  cudaError_t res = cudaGetLastError();\
  if (res != cudaSuccess)\
    {\
      fprintf(stderr, "Line %d: CUDA Error: %s\n", \
              __LINE__, cudaGetErrorString(res));\
      cudaThreadExit();\
      exit(1);\
    }\
}



void readAndCompileShaderFromGLSLFile(GLuint new_shaderprogram, const char *filename, GLenum shaderType)
{
  FILE *file = fopen(filename,"rb"); // open shader text file
  if (!file) 
    error_exit("Filename %s does not exist\n", filename);

  /* get the size of the file and read it */
  fseek(file,0,SEEK_END);
  GLint size = ftell(file);
  char *data = (char*)malloc(sizeof(char)*(size + 1));
  memset(data, 0, sizeof(char)*(size + 1));
  fseek(file,0,SEEK_SET);
  size_t res = fread(data,1,size,file);
  fclose(file);
  
  GLuint shader = glCreateShader(shaderType);
  glShaderSource(shader, 1, (const GLchar**)&data, &size);
  glCompileShader(shader);

  GET_GLERROR(0);
  GLint compile_success = 0;
  glGetShaderiv(shader, GL_COMPILE_STATUS, &compile_success);
  GET_GLERROR(0);
  
  if (compile_success == GL_FALSE)
    {
      printf("Compilation of %s failed!\n Reason:\n", filename);

      GLint maxLength = 0;
      glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &maxLength);
      
      char errorLog[maxLength];
      glGetShaderInfoLog(shader, maxLength, &maxLength, &errorLog[0]);
      
      printf("%s", errorLog);
      
      glDeleteShader(shader); 
      exit(1);
    }

  glAttachShader(new_shaderprogram, shader);
  glDeleteShader(shader); // good to do? 

  free(data);  
}

GLuint ShaderCreate(const char *vshader_filename, const char *fshader_filename)
{
  printf("Loading GLSL shaders %s %s\n", vshader_filename, fshader_filename);

  GLuint new_shaderprogram = glCreateProgram();

  GET_GLERROR(0);
  if (vshader_filename)
    readAndCompileShaderFromGLSLFile(new_shaderprogram, vshader_filename, GL_VERTEX_SHADER);

  GET_GLERROR(0);
  if (fshader_filename)
    readAndCompileShaderFromGLSLFile(new_shaderprogram, fshader_filename, GL_FRAGMENT_SHADER);
  
  GET_GLERROR(0);
  
  glLinkProgram(new_shaderprogram);
  
  GET_GLERROR(0);
  GLint link_success;
  glGetProgramiv(new_shaderprogram, GL_LINK_STATUS, &link_success);

  if (link_success == GL_FALSE)
    {
      printf("Linking of %s with %s failed!\n Reason:\n", vshader_filename, fshader_filename);

      GLint maxLength = 0;
      glGetShaderiv(new_shaderprogram, GL_INFO_LOG_LENGTH, &maxLength);
      
      char errorLog[maxLength];
      glGetShaderInfoLog(new_shaderprogram, maxLength, &maxLength, &errorLog[0]);
      
      printf("%s", errorLog);
      
      exit(EXIT_FAILURE);
    }

  return new_shaderprogram;
}



void CudaGLES::InitCudaGLES(int width, int height, char *window_name)
{
  
  graphics_setup_window(0,0, width, height, (const char *)window_name);	
	
  char *GL_version=(char *)glGetString(GL_VERSION);
  char *GL_vendor=(char *)glGetString(GL_VENDOR);
  char *GL_renderer=(char *)glGetString(GL_RENDERER);
  
  printf("Version: %s\n", GL_version);
  printf("Vendor: %s\n", GL_vendor);
  printf("Renderer: %s\n", GL_renderer);

  m_width = width;
  m_height = height;

  
    // initialize buffer object
  glGenBuffers(1, &m_vbo);
  glBindBuffer(GL_ARRAY_BUFFER, m_vbo);
	

  unsigned int size = width * height * 4 * sizeof(float);
  glBufferData(GL_ARRAY_BUFFER, size, NULL, GL_DYNAMIC_DRAW);
  glVertexAttribPointer((GLuint)0, 4, GL_FLOAT, GL_FALSE, 0, 0);
  glEnableVertexAttribArray(0); 
  
  cudaGraphicsGLRegisterBuffer(&m_position, m_vbo, cudaGraphicsMapFlagsNone);
  
  glGenBuffers(1, &m_vbo);
  glBindBuffer(GL_ARRAY_BUFFER, m_vbo);


  size = width * height * 4 * sizeof(float);
  glBufferData(GL_ARRAY_BUFFER, size, NULL, GL_DYNAMIC_DRAW);
  glVertexAttribPointer((GLuint)1, 4, GL_FLOAT, GL_FALSE, 0, 0);
  glEnableVertexAttribArray(1); 
  
  cudaGraphicsGLRegisterBuffer(&m_color, m_vbo, cudaGraphicsMapFlagsNone);

  // GLSL stuff
  char *vertex_shader_path = sdkFindFilePath("/home/ubuntu/TX1-Pano/gles_render/mesh.vert.glsl", NULL);
  char *fragment_shader_path = sdkFindFilePath("/home/ubuntu/TX1-Pano/gles_render/mesh.frag.glsl", NULL);

  if (vertex_shader_path == NULL || fragment_shader_path == NULL)
  {
    printf("Error finding shader file\n");
    exit(EXIT_FAILURE);
  }

  m_shader = ShaderCreate(vertex_shader_path, fragment_shader_path);
//  GET_GLERROR(0);

  free(vertex_shader_path);
  free(fragment_shader_path);
  
  glUseProgram(m_shader);
  
  glClear( GL_COLOR_BUFFER_BIT );
  graphics_swap_buffers();
  
}
	
	
	
	
	






__global__ void simple_vbo_kernel(float4 *pos, uchar4 *cuda_input, float4 *color, unsigned int width, unsigned int height)
{
    unsigned int x = blockIdx.x*blockDim.x + threadIdx.x;
    unsigned int y = blockIdx.y*blockDim.y + threadIdx.y;

    // calculate uv coordinates
    float u = x / (float) width;
    float v = y / (float) height;
    u = u*2.0f - 1.0f;
    v = v*2.0f - 1.0f;

	cuda_input += (y*width + x);
	
	float r = (*cuda_input).x;
	float g = (*cuda_input).y;
	float b = (*cuda_input).z;
	
	r/=255.0;
	g/=255.0;
	b/=255.0;
	
    // write output vertex
    pos[y*width+x] = make_float4(u, v, 0, 1.0f);
    color[y*width+x] = make_float4(r, g, b, 1.0f);
}


void launch_kernel(float4 *pos, uchar4 *cuda_input, float4 * color, unsigned int mesh_width,
                   unsigned int mesh_height)
{
    // execute the kernel
    dim3 block(8, 8, 1);
    dim3 grid(mesh_width / block.x, mesh_height / block.y, 1);
    simple_vbo_kernel<<< grid, block>>>(pos, cuda_input, color, mesh_width, mesh_height);
}

void runCuda(struct cudaGraphicsResource **position_resource, struct cudaGraphicsResource **color_resource, uchar4 *rgba_buff, int width, int height)
{
    // map OpenGL buffer object for writing from CUDA
    float4 *dptr, *color;
 
    
    cudaGraphicsMapResources(1, position_resource, 0);
    cudaGraphicsMapResources(1, color_resource, 0);
    size_t num_bytes;
    cudaGraphicsResourceGetMappedPointer((void **)&dptr, &num_bytes, *position_resource);
    cudaGraphicsResourceGetMappedPointer((void **)&color, &num_bytes, *color_resource);
   
	
	launch_kernel(dptr, rgba_buff, color, width, height);

    cudaGraphicsUnmapResources(1, position_resource, 0);
    cudaGraphicsUnmapResources(1, color_resource, 0);
}

GLuint mesh_shader = 0;



void CudaGLES::RenderCudaBuffer(unsigned char *buff)
{
	uchar4 *rgba_buff = reinterpret_cast<uchar4 *>(buff);
    

    // run CUDA kernel to generate vertex  positions and color
    runCuda(&m_position, &m_color, rgba_buff, m_width, m_height);

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
    glDrawArrays(GL_POINTS, 0, m_width * m_height);

    glFinish();
    
    usleep(1000);
			
	graphics_swap_buffers();
}
void CudaGLES::stop()
{
	cudaGraphicsUnregisterResource(m_position); 
	cudaGraphicsUnregisterResource(m_color);      
	
	graphics_close_window(); // close window and destroy OpenGL ES context
}
/*
int main()
{
	int width = 1920;
	int height = 1080;
	
	unsigned char *h, *d;
	h = (unsigned char *)malloc(width * height * 4);
	cudaMalloc((void **)&d, width*height*4);
	
	FILE *f;
	f = fopen("input.yuv", "r");
	fread(h, width*height*4, 1, f);
	fclose(f);
	cudaMemcpy(d, h, width*height*4, cudaMemcpyHostToDevice);
	
	CudaGLES obj;
	obj.InitCudaGLES(width, height, "Render Pano");
	for(int i = 0 ;  i < 200 ; i++)
	{
		obj.RenderCudaBuffer(d);
		//usleep(1000);
	}
	return 0;
}*/
